# Copyright 2014-2016 Nervana Systems Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Definition of the GPU layers
These layers are mainly used for old benchmarking code,
but they also cache all the computed params for complex layers.
TODO: clean up merge with CPU layers
TODO: remove any non-param caching code, neon layers should replace benchmark code.
"""
from __future__ import division
from builtins import object, str
import logging
from math import ceil
import numpy as np
from operator import mul
import pycuda.driver as drv
from pycuda.tools import context_dependent_memoize
import sys
from neon import logger as neon_logger
from neon.backends import kernel_specs
from neon.backends import convolution

logger = logging.getLogger(__name__)


if sys.version_info >= (3, 0):
    from functools import reduce


class Layer(object):

    """
    GPU Layer base class
    """

    def __init__(self, lib, dtype, N, dtypeU=None):

        if hasattr(dtype, 'type'):
            self.dtype = dtype
        else:
            self.dtype = np.dtype(dtype)

        self.N      = N
        self.dtypeU = dtype if dtypeU is None else dtypeU
        self.lib    = lib
        self.flops  = 0
        self.sizeI  = 0
        self.sizeO  = 0
        self.sizeF  = 0
        self.weights   = None
        self.fprop_in  = None
        self.fprop_out = None
        self.bprop_in  = None
        self.bprop_out = None

        self.learning_rate = 0.0

    def init_activations(self, fprop_out=None):

        if fprop_out is not None:
            self.fprop_out = fprop_out
        else:
            self.fprop_out = self.lib.empty(self.dimO, dtype=self.dtype)

        self.act_stats = self.lib.empty((self.dimO2[0], 1), dtype=np.float32)

    def init_deltas(self, shared=None):

        if shared is None:
            self.bprop_out = self.lib.empty(self.dimI, dtype=self.dtype)
        else:
            self.bprop_out = shared[0].share(self.dimI)
            shared.reverse()

        self.delta_stats = self.lib.empty((self.dimI2[0], 1), dtype=np.float32)

    def init_weights(self, loc=0.0, scale=0.1, shared=None, zeros=False):

        if self.sizeF > 0:
            if zeros:
                self.weights  = self.lib.zeros(self.dimF, dtype=self.dtype)
            else:
                weights       = np.random.normal(loc, scale, self.dimF)
                self.weights  = self.lib.array(weights, dtype=self.dtype)

            if shared is None:
                self.updat_out = self.lib.empty(self.dimF, dtype=self.dtypeU)
            else:
                self.updat_out = shared.share(self.dimF, dtype=self.dtypeU)

            self.weight_stats = self.lib.empty((self.dimF2[0], 1), dtype=np.float32)

    def scale_weights(self, scale):

        mean = self.get_activation_mean()
        self.weights[:] *= scale / mean

    def fprop(self, fprop_in, scale_weights=0):
        if self.fprop_in is None and fprop_in:
            self.fprop_in = fprop_in.reshape(self.dimI)
        return self.fprop_in

    def bprop(self, bprop_in, beta=0):
        return bprop_in

    # fprop relu happens inside of the conv and gemm kernels
    def bprop_relu(self, bprop_in):

        bprop_in[:] = bprop_in * (self.fprop_out > 0)
        return bprop_in

    def grad_descent(self):
        self.weights[:] += self.updat_out * self.learning_rate

    def get_activation_mean(self):
        return self._get_mean(self.fprop_out, self.act_stats, self.dimO2)

    def get_activation_max(self):
        return self._get_max(self.fprop_out, self.act_stats, self.dimO2)

    def get_delta_mean(self):
        return self._get_mean(self.bprop_out, self.delta_stats, self.dimI2)

    def get_delta_max(self):
        return self._get_max(self.bprop_out, self.delta_stats, self.dimI2)

    def get_update_mean(self):
        if self.sizeF > 0:
            return self._get_mean(self.updat_out, self.weight_stats, self.dimF2)
        return 0

    def get_update_max(self):
        if self.sizeF > 0:
            return self._get_max(self.updat_out, self.weight_stats, self.dimF2)
        return 0

    def get_weight_mean(self):
        if self.sizeF > 0:
            return self._get_mean(self.weights, self.weight_stats, self.dimF2)
        return 0

    def get_weight_max(self):
        if self.sizeF > 0:
            return self._get_max(self.weights, self.weight_stats, self.dimF2)
        return 0

    def _get_mean(self, ary, buf, shape):

        buf1    = buf[0:1, 0:1]
        buf[:]  = self.lib.sum(abs(ary.reshape(shape)), axis=1)
        buf1[:] = self.lib.sum(buf, axis=0) * (1.0 / ary.size)
        return float(buf1.get()[0, 0])

    def _get_max(self, ary, buf, shape):

        buf1    = buf[0:1, 0:1]
        buf[:]  = self.lib.max(abs(ary.reshape(shape)), axis=1)
        buf1[:] = self.lib.max(buf, axis=0)
        return float(buf1.get()[0, 0])

    def fprop_stats(self):
        neon_logger.display("fprop:%10.5f mean %11.5f max %s"
              % (self.get_activation_mean(), self.get_activation_max(), self))

    def bprop_stats(self):
        if self.bprop_out is not None:
            neon_logger.display("bprop:%10.5f mean %11.5f max %s"
                  % (self.get_delta_mean(), self.get_delta_max(), self))

        if self.weights is not None:
            up_mean, up_max = (self.get_update_mean(), self.get_update_max())
            wt_mean, wt_max = (self.get_weight_mean(), self.get_weight_max())
            rt_mean, rt_max = (0.0001 * up_mean / wt_mean, 0.0001 * up_max / wt_max)
            neon_logger.display("updat:%10.5f mean %11.5f max %s" % (up_mean, up_max, self))
            neon_logger.display("weigh:%10.5f mean %11.5f max" % (wt_mean, wt_max))
            neon_logger.display("ratio:%10.5f mean %11.5f max" % (rt_mean, rt_max))

    @staticmethod
    def create(lib, conf, prev_layer, dtype):

        config     = dict(conf)
        layer_type = config.pop("layer")

        # merge dtype specific settings
        config["dtype"] = dtype

        # merge shared params
        config.update(config.pop("common", {}))

        # Propagate the fixed and calculated dimensions
        if prev_layer is not None:
            config["N"] = prev_layer.N

            if layer_type is FullLayer:
                config["nIn"] = prev_layer.nOut
            elif layer_type is PoolLayer and type(prev_layer) is FullLayer:
                config["C"] = prev_layer.nOut
            elif layer_type is BatchNorm and type(prev_layer) is FullLayer:
                config["nIn"] = prev_layer.nOut
            else:
                config["C"] = prev_layer.K
                config["D"] = prev_layer.M
                config["H"] = prev_layer.P
                config["W"] = prev_layer.Q

                if layer_type is Inception:
                    partitions  = config.pop("partitions")
                    config["K"] = 0

                    config["partitions"] = []
                    for part in partitions:
                        layer_sequence = []
                        part_prev_layer = prev_layer
                        for layer_conf in part:
                            part_prev_layer = Layer.create(lib, layer_conf, part_prev_layer, dtype)
                            layer_sequence.append(part_prev_layer)

                        last = layer_sequence[-1]
                        config["partitions"].append(layer_sequence)
                        config["K"] += last.K
                        if "P" in config:
                            assert config["P"] == last.P and config["Q"] == last.Q
                        else:
                            config["M"] = last.M
                            config["P"] = last.P
                            config["Q"] = last.Q

        # Instantiate the layer
        return layer_type(lib, **config)


class DataLayer(Layer):

    """
    Input data layer.
    """

    def __init__(self, lib, dtype, N, C, D=1, H=1, W=1):

        super(DataLayer, self).__init__(lib, dtype, N)

        self.C = C
        self.K = C
        self.M = D
        self.P = H
        self.Q = W
        self.DHW = (D, H, W)
        self.dimI   = (C, D, H, W, N)
        self.dimO   = (C, D, H, W, N)
        self.dimI2  = (C * D * H * W, N)
        self.dimO2  = (C * D * H * W, N)
        self.sizeO  = reduce(mul, self.dimO, 1)
        self.sizeI  = self.sizeO

    def init_data(self, ary=None):
        if ary is None:
            self.fprop_out.fill(0)
        else:
            self.fprop_out.set(ary)

    def init_deltas(self, shared=None):
        pass

    def init_weights(self, loc=0.0, scale=0.1, shared=None, zeros=False):
        pass

    def fprop(self, fprop_in, scale_weights=0):
        return self.fprop_out

    def __str__(self):
        return "DataLayer: NCK: (%d, %d, %d) DHW:%s" % (self.N, self.C, self.K, self.DHW)


class FullLayer(Layer):

    """
    Fully connnected layer.
    """

    def __init__(self, lib, dtype, N, nIn, nOut, relu=False):

        super(FullLayer, self).__init__(lib, dtype, N)

        self.nIn    = nIn
        self.nOut   = nOut
        self.flops  = N * nIn * nOut * 2.0
        self.dimI   = (nIn, N)
        self.dimI2  = (nIn, N)
        self.dimO   = (nOut, N)
        self.dimO2  = (nOut, N)
        self.dimF   = (nOut, nIn)
        self.dimF2  = (nOut, nIn)
        self.sizeI  = nIn  * N
        self.sizeO  = nOut * N
        self.sizeF  = nIn  * nOut
        self.relu   = relu

    def fprop(self, fprop_in, scale_weights=0):

        fprop_in = super(FullLayer, self).fprop(fprop_in)
        self.lib.compound_dot(self.weights, fprop_in, self.fprop_out, self.relu)

        if scale_weights:
            self.scale_weights(scale_weights)
            self.fprop(fprop_in)

        return self.fprop_out

    def bprop(self, bprop_in, beta=0):

        if self.relu:
            self.bprop_relu(bprop_in)
        self.lib.compound_dot(self.weights.T, bprop_in, self.bprop_out)

        self.lib.compound_dot(bprop_in, self.fprop_in.T, self.updat_out)
        self.grad_descent()

        return self.bprop_out

    def __str__(self):
        return "FullLayer: N, nIn, nOut: (%d, %d, %d)" % (self.N, self.nIn, self.nOut)


class ConvLayer(Layer):

    """
    ConvLayer parameter object.
    This then is passed as an argument to all the convolution operations.

    N: Number of images in mini-batch
    C: Number of input feature maps
    K: Number of output feature maps

    D: Depth  of input image
    H: Height of input image
    W: Width  of input image

    T: Depth  of filter kernel
    R: Height of filter kernel
    S: Width  of filter kernel

    padding: amount of zero-padding around the given edge
    strides: factor to step the filters by in a given direction
    dilation: dilation factor for each dimension
    """

    def __init__(self, lib, dtype,
                 N, C, K,
                 D=1, H=1, W=1,
                 T=1, R=1, S=1,
                 pad_d=0, pad_h=0, pad_w=0,
                 str_d=1, str_h=1, str_w=1,
                 dil_d=1, dil_h=1, dil_w=1):

        super(ConvLayer, self).__init__(lib, dtype, N, np.float32)

        # Compute the output spatial dimensions
        M = lib.output_dim(D, T, pad_d, str_d, pooling=False, dilation=dil_d)
        P = lib.output_dim(H, R, pad_h, str_h, pooling=False, dilation=dil_h)
        Q = lib.output_dim(W, S, pad_w, str_w, pooling=False, dilation=dil_w)

        self.C = C
        self.K = K
        self.M = M
        self.P = P
        self.Q = Q
        self.NCK = (N, C, K)
        self.TRS = (T, R, S)
        self.DHW = (D, H, W)
        self.MPQ = (M, P, Q)
        self.padding = (pad_d, pad_h, pad_w)
        self.strides = (str_d, str_h, str_w)

        self.all_params = (N, C, K, D, H, W, T, R, S, pad_d, pad_h, pad_w, str_d, str_h, str_w,
                           dil_d, dil_h, dil_w)

        self.dimI   = (C, D, H, W, N)
        self.dimF   = (C, T, R, S, K)
        self.dimFb  = (K, T, R, S, C)
        self.dimO   = (K, M, P, Q, N)
        self.dimI2  = (C * D * H * W, N)
        self.dimF2  = (C * T * R * S, K)
        self.dimF2t = (K, C * T * R * S)
        self.dimO2  = (K * M * P * Q, N)
        self.dimS   = (K, 1)
        self.sizeI  = reduce(mul, self.dimI, 1)
        self.sizeF  = reduce(mul, self.dimF, 1)
        self.sizeO  = reduce(mul, self.dimO, 1)
        self.nOut   = reduce(mul, self.MPQ, 1) * K

        # flop count for benchmarking
        self.flops = P * Q * M * K * N * C * R * S * T * 2.0

        args = (lib, self.dtype, N, C, K, D, H, W, T, R, S, M, P, Q,
                pad_d, pad_h, pad_w, str_d, str_h, str_w, dil_d, dil_h, dil_w)

        #lib.enable_winograd = 0

        dilated_conv = (dil_d != 1 or dil_h != 1 or dil_w != 1)
        if dilated_conv:
            assert (dil_w > 0 and dil_h > 0 and dil_w > 0)

        ####### Cuda C ###########
        if lib.use_cudac_kernels:

            #3D conv not supported yet
            if T > 1 or D > 1:
                raise ValueError("3D Convolution not supported by CUDA C kernels and pre-Maxwell GPUs")

            # TODO small C bprop?
            self.fprop_kernels = convolution.FpropCuda(*args)
            self.bprop_kernels = convolution.BpropCuda(*args)
            self.updat_kernels = convolution.UpdateCuda(*args)

        ####### Winograd ###########
        elif lib.enable_winograd and R == 3 and S == 3 and all(x == 1 for x in (D,M,T,str_w,str_h,str_d)) and not dilated_conv:
            from .winograd_conv import (FpropWinograd_2x2_3x3, BpropWinograd_2x2_3x3, UpdateWinograd_3x3_2x2,
                                        FpropWinograd_4x4_3x3, BpropWinograd_4x4_3x3, UpdateWinograd_3x3_4x4)

            # Temp for now till we can autotune
            # 2 is safer for fp16 without batchnorm
            if dtype == np.float32 and lib.enable_winograd == 4:
                winograd = 4
            else:
                winograd = 2

            if C < 8:
                self.fprop_kernels = convolution.FpropDirect(*args)
            elif winograd == 4 and H * W < 112 * 112:
                self.fprop_kernels = FpropWinograd_4x4_3x3(*args)
            else:
                self.fprop_kernels = FpropWinograd_2x2_3x3(*args)

            if winograd == 4 and H * W < 112 * 112:
                self.bprop_kernels = BpropWinograd_4x4_3x3(*args)
            else:
                self.bprop_kernels = BpropWinograd_2x2_3x3(*args)

            if N >= 4 and (C < 8 or H * W > 112 * 112):
                self.updat_kernels = convolution.UpdateDirect(*args)
            elif winograd == 4:
                self.updat_kernels = UpdateWinograd_3x3_4x4(*args)
            else:
                self.updat_kernels = UpdateWinograd_3x3_2x2(*args)

#        elif lib.enable_winograd and not lib.deterministic and N > 1 and \
#            R == 5 and S == 5 and all(x == 1 for x in (D,M,T,str_w,str_h,str_d)):
#
#            from .winograd_conv import (FpropWinograd_2x2_5x5, BpropWinograd_2x2_5x5)
#
#            self.fprop_kernels = FpropWinograd_2x2_5x5(*args)
#            self.bprop_kernels = BpropWinograd_2x2_5x5(*args)
#            if N >= 4:
#                self.updat_kernels = convolution.UpdateDirect(*args)

        ####### Direct ###########
        else:

            self.fprop_kernels = convolution.FpropDirect(*args)
            self.bprop_kernels = convolution.BpropDirect(*args)
            if N >= 4:
                self.updat_kernels = convolution.UpdateDirect(*args)

        #logger.debug("%s: %s, %s, %s", str(self), str(self.fprop_kernels), str(self.bprop_kernels), str(self.updat_kernels))


    def init_activations(self, fprop_out=None):

        super(ConvLayer, self).init_activations(fprop_out)

        if self.bsum:
            self.batch_sum = self.lib.empty(self.dimS, dtype=np.float32)
        else:
            self.batch_sum = None

    def fprop(self, fprop_in, scale_weights=0):
        """
        Conv Layer forward propagation.

        Arguments:
            fprop_in (Tensor): Inputs
            scale_weights (float): Scale weights by scale/mean if nonzero

        Returns:
            fprop_out (Tensor): Output activations
        or
            (self.fprop_out, self.batch_sum) (tuple): Tuple with batch_sum
                added as the second entry.
        """
        fprop_in = super(ConvLayer, self).fprop(fprop_in)
        self.lib.fprop_conv(self, fprop_in, self.weights, self.fprop_out, bsum=self.batch_sum)

        if scale_weights:
            self.scale_weights(scale_weights)
            self.fprop(fprop_in)

        if self.bsum:
            return (self.fprop_out, self.batch_sum)
        return self.fprop_out

    def bprop(self, bprop_in, beta=0):

        if self.relu:
            self.bprop_relu(bprop_in)
        if self.bprop_out is not None:
            self.lib.bprop_conv(self, self.weights, bprop_in, self.bprop_out, beta=beta)

        self.lib.update_conv(self, self.fprop_in, bprop_in, self.updat_out)
        self.grad_descent()

        return self.bprop_out

    def __str__(self):
        return ("ConvLayer: NCK: (%3d, %3d, %3d) HW:%s" %
                (self.N, self.C, self.K, self.DHW[1:3]))

# Add Deconv class
class DeconvLayer(ConvLayer):

    """
    DeconvLayer parameter object.
    This then is passed as an argument to all the convolution operations.

    N: Number of images in mini-batch
    C: Number of output feature maps
    K: Number of input feature maps

    M: Depth of input
    P: Height of input
    Q: Width of input

    D: Depth  of output image
    H: Height of output image
    W: Width  of output image

    T: Depth  of filter kernel
    R: Height of filter kernel
    S: Width  of filter kernel

    padding: amount of zero-padding around the given edge
    strides: factor to step the filters by in a given direction
    dilation: dilation factor for each dimension
    """

    def __init__(self, lib, dtype,
                 N, C, K,
                 M, P, Q,
                 T=1, R=1, S=1,
                 pad_d=0, pad_h=0, pad_w=0,
                 str_d=1, str_h=1, str_w=1,
                 dil_d=1, dil_h=1, dil_w=1):

        tt = dil_d * (T - 1) + 1
        rr = dil_h * (R - 1) + 1
        ss = dil_w * (S - 1) + 1
        # Cannot get exact, e.g. because not unique
        D = (M - 1) * str_d - 2 * pad_d + tt
        H = (P - 1) * str_h - 2 * pad_h + rr
        W = (Q - 1) * str_w - 2 * pad_w + ss

        super(DeconvLayer, self).__init__(
            lib, dtype,
            N, C, K,
            D, H, W,
            T, R, S,
            pad_d, pad_h, pad_w,
            str_d, str_h, str_w,
            dil_d, dil_h, dil_w)

        self.nOut = reduce(mul, self.DHW, 1) * C

    def __str__(self):
        return ("DeconvLayer: NCK: (%d, %d, %d) DHW:%s TRS:%s MPQ:%s" %
                (self.N, self.C, self.K, self.DHW, self.TRS, self.MPQ))


class PoolLayer(Layer):

    """
    PoolLayer parameter object.
    This then is passed as an argument to all pooling kernels.

    op: max, avg, l2 pooling
    N: Number of images in mini-batch

    C: Number of input feature maps
    D: Depth  of input image
    H: Height of input image
    W: Width  of input image

    J: Size of feature map pooling window (maxout n_pieces)
    T: Depth  of pooling window
    R: Height of pooling window
    S: Width  of pooling window

    padding: amount of zero-padding around the given image or feature map edge
    strides: factor to step the window by in a given direction (overlap allowed)

    Leave spatial dimensions at 1 to allow feature map pooling in the fc layers.

    """

    def __init__(self, lib, dtype,
                 op, N, C,
                 D=1, H=1, W=1,
                 J=1, T=1, R=1, S=1,
                 pad_c=0, pad_d=0, pad_h=0, pad_w=0,
                 str_c=None, str_d=None, str_h=None, str_w=None):

        super(PoolLayer, self).__init__(lib, dtype, N)

        if self.dtype.type is np.float16:
            clss = "hpool"
        elif self.dtype.type is np.float32:
            clss = "spool"
        else:
            raise TypeError("Type not supported.")

        # default to non-overlapping
        if str_c is None:
            str_c = J
        if str_d is None:
            str_d = T
        if str_h is None:
            str_h = R
        if str_w is None:
            str_w = S

#        if str_c < J or str_d < T or str_h < R or str_w < S:
#            self.overlap = (ceil(float(J) / str_c) *
#                            ceil(float(T) / str_d) *
#                            ceil(float(R) / str_h) *
#                            ceil(float(S) / str_w))
#        else:
#            self.overlap = 0.0

        self.overlap = 1.0

        # TODO: detect other forms of gaps
        if str_c > J or str_d > T or str_h > R or str_w > S:
            self.gaps = 1
        else:
            self.gaps = 0

        bprop_zero = self.overlap or self.gaps

        # Compute the output dimensions
        K = lib.output_dim(C, J, pad_c, str_c, pooling=True)
        M = lib.output_dim(D, T, pad_d, str_d, pooling=True)
        P = lib.output_dim(H, R, pad_h, str_h, pooling=True)
        Q = lib.output_dim(W, S, pad_w, str_w, pooling=True)

        self.op   = op
        self.C    = C
        self.K    = K
        self.M    = M
        self.P    = P
        self.Q    = Q
        self.JTRS = (J, T, R, S)
        self.DHW  = (D, H, W)
        self.MPQ  = (M, P, Q)
        self.padding = (pad_c, pad_d, pad_h, pad_w)
        self.strides = (str_c, str_d, str_h, str_w)

        self.dimI   = (C, D, H, W, N)
        self.dimO   = (K, M, P, Q, N)
        self.dimF2  = None
        self.dimI2  = (C * D * H * W, N)
        self.dimO2  = (K * M * P * Q, N)
        self.sizeI  = reduce(mul, self.dimI, 1)
        self.sizeO  = reduce(mul, self.dimO, 1)
        self.nOut   = reduce(mul, self.MPQ, 1) * K

        # precompute some multiplications for fast constant memory access
        WN   = W * N
        HWN  = H * WN
        DHWN = D * HWN
        DH   = D * H
        RS   = R * S
        RST  = T * RS
        JRST = J * RST
        QN   = Q * N
        PQN  = P * QN
        MPQN = M * PQN

        assert JRST + 32 < 2**16, "Integer division is faster with 16bit numerators"

        sb_large = {
            #SB  shlP maskP shrP shlQ maskQ shrQ maskN shrN
            1  : (0,   0x00, 0,   0,   0x00, 0,   0xfff, 32), # 1x1  nnnnn
            2  : (0,   0x00, 0,   1,   0x10, 4,   0x00f,  4), # 1x2  xnnnn
            4  : (0,   0x00, 0,   2,   0x18, 3,   0x007,  3), # 1x4  xxnnn
            8  : (0,   0x00, 0,   3,   0x1c, 2,   0x003,  2), # 1x8  xxxnn
            16 : (0,   0x00, 0,   4,   0x1e, 1,   0x001,  1), # 1x16 xxxxn
            32 : (0,   0x00, 0,   5,   0x1f, 0,   0x000,  0), # 1x32 xxxxx
        }
        sb_medium = {
            #SB  shlP maskP shrP shlQ maskQ shrQ maskN shrN
            8  : (1,   0x10, 4,   2,   0x0c, 2,   0x003,  2), # 2x4  yxxnn
            16 : (1,   0x10, 4,   3,   0x0e, 1,   0x001,  1), # 2x8  yxxxn
            32 : (1,   0x10, 4,   4,   0x0f, 0,   0x000,  0), # 2x16 yxxxx
        }
        sb_small = {
            #SB  shlP maskP shrP shlQ maskQ shrQ maskN shrN
            16 : (2,   0x18, 3,   2,   0x06, 1,   0x001,  1), # 4x4  yyxxn
            32 : (2,   0x18, 3,   3,   0x07, 0,   0x000,  0), # 4x8  yyxxx
        }

        if N == 1:
            super_block = 0
        elif N < 32:
            super_block = len(bin(N - 1)) - 2
        else:
            super_block = 5
        super_block = 1 << (5 - super_block)

        # try to minimize the zero overlap in the superblock
        # but maximize the x dim of the superblock for more contiguous memory access
        if super_block < 8 or Q > 64:
            sb_params = sb_large.get(super_block)
        elif super_block < 16 or Q > 32:
            sb_params = sb_medium.get(super_block)
        else:
            sb_params = sb_small.get(super_block)

        supP = _ceil_div(P, 1 << sb_params[0])
        supQ = _ceil_div(Q, 1 << sb_params[3])

        # precompute the magic numbers and shift amounts for integer division
        magic_RST = _magic32(JRST + 32, RST)
        magic_RS  = _magic32(RST + 32, RS)
        magic_S   = _magic32(RS + 32, S)
        magic_P   = _magic32(M * supP, supP)

        fprop_name = "fprop_" + op
        bprop_name = "bprop_" + op

        threads = 32 if super_block > 1 else N

        self.fprop_kernel = [fprop_name, (supQ, supP * M, K), (threads, 1, 1), _flatten([
            N, W, H, D, C, WN, HWN, DHWN,
            P, Q, magic_P, QN, PQN, MPQN,
            pad_c, pad_d, pad_h, pad_w,
            str_c, str_d, str_h, str_w,
            S, RS, RST, JRST, magic_S, magic_RS, magic_RST,
            supP, supQ, sb_params ])]

        lut_size = JRST
        if lut_size % 4 != 0:
            lut_size += 4 - lut_size % 4

        self.bprop_lut_size = self.fprop_lut_size = super_block * lut_size * 4

        if self.overlap > 0:

            # we have a special kernel to handle the overlapping avg pooling
            bprop_name += "_overlap"

            magic_str_w = _magic32(W + S, str_w)
            magic_str_h = _magic32(H + R, str_h)
            magic_str_d = _magic32(D + T, str_d)
            magic_str_c = _magic32(C + J, str_c)

            if super_block > 1:

                bprop_name += "_smallN"

                if super_block < 8 or W > 64:
                    sb_params = sb_large.get(super_block)
                elif super_block < 16 or W > 32:
                    sb_params = sb_medium.get(super_block)
                else:
                    sb_params = sb_small.get(super_block)

                supH = _ceil_div(H, 1 << sb_params[0])
                supW = _ceil_div(W, 1 << sb_params[3])

                magic_H = _magic32(D * supH, supH)

                maxLutSize = \
                    _ceil_div(S, str_w) * \
                    _ceil_div(R, str_h) * \
                    _ceil_div(T, str_d) * \
                    _ceil_div(J, str_c)

                #neon_logger.display((supW, D*supH, C), sb_params, maxLutSize)

                self.bprop_kernel = [bprop_name, (supW, D * supH, C), (threads, 1, 1), _flatten([
                    N, W, H, D, C, WN, HWN, DHWN, magic_H,
                    pad_w, pad_h, pad_d, pad_c,
                    str_w, str_h, str_d, str_c,
                    magic_str_w, magic_str_h, magic_str_d, magic_str_c,
                    S, R, T, J, RS, RST, JRST, magic_S, magic_RS, magic_RST,
                    Q, P, M, K, QN, PQN, MPQN,
                    supH, supW, sb_params, maxLutSize])]

                lut_size = maxLutSize
                if lut_size % 4 != 0:
                    lut_size += 4 - lut_size % 4

                self.bprop_lut_size = super_block * lut_size * 4 * 2

            else:

                # The overlap kernel can be much more efficient if we aren't doing superblocking
                magic_H = _magic32(DH, H)

                self.bprop_kernel = [bprop_name, (W, DH, C), (threads, 1, 1), _flatten([
                    N, W, H, D, C, WN, HWN, DHWN, magic_H,
                    pad_w, pad_h, pad_d, pad_c,
                    str_w, str_h, str_d, str_c,
                    magic_str_w, magic_str_h, magic_str_d, magic_str_c,
                    S, R, T, J, RS, RST, JRST, magic_S, magic_RS, magic_RST,
                    Q, P, M, K, QN, PQN, MPQN])]

                self.bprop_lut_size = lut_size * 4 * 2
        else:
            self.bprop_kernel = [bprop_name, (supQ, supP * M, K), (threads, 1, 1), _flatten([
                N, W, H, D, C, WN, HWN, DHWN,
                P, Q, magic_P, QN, PQN, MPQN,
                pad_c, pad_d, pad_h, pad_w,
                str_c, str_d, str_h, str_w,
                S, RS, RST, JRST, magic_S, magic_RS, magic_RST,
                supP, supQ, sb_params])]

    def init_activations(self, fprop_out=None):

        super(PoolLayer, self).init_activations(fprop_out)

        self.argmax = self.lib.empty(self.dimO, dtype=np.uint8)

    def fprop(self, fprop_in, scale_weights=0):
        """ Used for benchmarking only"""
        fprop_in = super(PoolLayer, self).fprop(fprop_in)
        self.lib.fprop_pool(self, fprop_in, self.fprop_out, argmax=self.argmax)
        return self.fprop_out

    def bprop(self, bprop_in, beta=0):
        self.lib.bprop_pool(self, bprop_in, self.bprop_out, argmax=self.argmax)
        return self.bprop_out

    def __str__(self):
        return ("PoolLayer: NCK: (%d, %d, %d) DHW:%s JTRS:%s MPQ:%s op: %s " %
                (self.N, self.C, self.K, self.DHW, self.JTRS, self.MPQ, self.op))


class Inception(Layer):

    """
    GoogLeNet inception assembly layer.

    """

    def __init__(self, lib, dtype, partitions,
                 N, C, K,
                 D=1, H=1, W=1,
                 M=1, P=1, Q=1):

        super(Inception, self).__init__(lib, dtype, N)

        self.partitions = partitions

        self.C = C
        self.K = K
        self.M = M
        self.P = P
        self.Q = Q
        self.NCK = (N, C, K)
        self.DHW = (D, H, W)
        self.MPQ = (M, P, Q)

        self.dimI  = (C, D, H, W, N)
        self.dimO  = (K, M, P, Q, N)
        self.dimI2 = (C * D * H * W, N)
        self.dimO2 = (K * M * P * Q, N)
        self.sizeI = reduce(mul, self.dimI, 1)
        self.sizeO = reduce(mul, self.dimO, 1)
        self.nOut  = reduce(mul, self.MPQ, 1) * K

        self.sizeF = 0
        self.flops = 0
        for part in partitions:
            for layer in part:
                self.flops += layer.flops
                self.sizeF  = max(self.sizeF, layer.sizeF)
                if self.sizeF == layer.sizeF:
                    self.dimF = layer.dimF

    def __str__(self):
        out = "Inception: NCK: (%d, %d, %d) DHW:%s MPQ:%s\n" % (self.N, self.C, self.K,
                                                                self.DHW, self.MPQ)
        for i, part in enumerate(self.partitions):
            out += "  Part%d:\n" % (i + 1)
            for layer in part:
                out += "    %s\n" % layer
        return out.rstrip()

    def init_activations(self, fprop_out=None):

        super(Inception, self).init_activations(fprop_out)
        K = 0
        for part in self.partitions:
            for layer in part:
                if layer is part[-1]:
                    layer.init_activations(self.fprop_out[K:K + layer.K, ...])
                    K += layer.K
                else:
                    layer.init_activations()

    def init_deltas(self, shared=None):

        super(Inception, self).init_deltas(shared)

        shared_deltas = shared[1:3] if shared else None

        for part in self.partitions:
            for layer in part:
                if layer is part[0]:
                    layer.bprop_out   = self.bprop_out
                    layer.delta_stats = self.delta_stats
                else:
                    layer.init_deltas(shared=shared_deltas)

    def init_weights(self, loc=0.0, scale=0.1, shared=None, zeros=False):

        for part in self.partitions:
            for layer in part:
                layer.init_weights(loc, scale, shared, zeros)

    def fprop(self, fprop_in, scale_weights=0):

        fprop_in = super(Inception, self).fprop(fprop_in)

        for part in self.partitions:
            part_fprop_in = fprop_in
            for layer in part:
                part_fprop_in = layer.fprop(part_fprop_in, scale_weights)

        return self.fprop_out

    def bprop(self, bprop_in):

        K = self.K
        for part in self.partitions[::-1]:
            part_bprop_in = bprop_in[K - part[-1].K:K, ...]
            K -= part[-1].K
            for layer in part[::-1]:
                if part is not self.partitions[-1] and layer is part[0]:
                    beta = 1.0
                else:
                    beta = 0.0

                part_bprop_in = layer.bprop(part_bprop_in, beta)

        return self.bprop_out

    def fprop_stats(self):
        for part in self.partitions:
            for layer in part:
                layer.fprop_stats()

    def bprop_stats(self):
        for part in self.partitions[::-1]:
            for layer in part[::-1]:
                layer.bprop_stats()


class BatchNorm(Layer):

    """
    Batch Normalization Layer
    """

    def __init__(self, lib, dtype, N, C=None, D=1, H=1, W=1, nIn=None, rho=0.99, eps=1e-6,
                 relu=False, bsum=False):
        """
        Batch Normalization layer

        Arguments:
            lib (Class): NervanaGPU instance
            dtype (Dtype): Data type
            N (int): batch size
            C (int): Number of input feature maps
            D (int): Depth  of input feature maps
            H (int): Height of input feature maps
            W (int): Width  of input feature maps
            nIn (int): Number on inputs for fully connected layer
            rho (float): Exponential window averaging factor
            eps (float): Constant added for numerical stability
            relu (bool): Flag for rectified linear activation function
            bsum (bool): PQN sum precomputed in conv kernel
        """
        super(BatchNorm, self).__init__(lib, dtype, N)

        self.rho  = rho
        self.eps  = eps
        self.relu = relu
        self.bsum = bsum

        if C is not None:

            self.C     = C
            self.K     = C
            self.M     = D
            self.P     = H
            self.Q     = W
            self.dimI  = (C, D, H, W, N)
            self.dimO  = (C, D, H, W, N)
            self.dimO2 = (C * D * H * W, N)
            self.dim2  = (C, D * H * W * N)
            self.nOut  = C * D * H * W

        elif nIn is not None:

            self.nOut  = nIn
            self.K     = nIn
            self.dimI  = (nIn, N)
            self.dimO  = (nIn, N)
            self.dimO2 = (nIn, N)
            self.dim2  = (nIn, N)

        else:
            raise ValueError("missing C or nIn")

        self.rcp_depth = 1.0 / self.dim2[1]

    def __str__(self):
        return ("BatchNorm: (%d, %d)" % self.dim2)

    def init_activations(self, fprop_out=None):

        if fprop_out is not None:
            self.fprop_out = fprop_out.reshape(self.dim2)
        else:
            self.fprop_out = self.lib.empty(self.dim2, dtype=self.dtype)

        self.xvar = self.lib.empty((self.K, 1), dtype=self.dtype)
        if not self.bsum:
            self.xsum = self.lib.empty((self.K, 1), dtype=np.float32)

    def init_deltas(self, shared=None):
        pass

    def init_weights(self, loc=0.0, scale=0.1, shared=None, zeros=False):

        lib = self.lib

        self.beta  = lib.zeros((self.K, 1), dtype=self.dtype)
        self.gamma = lib.ones((self.K, 1),  dtype=self.dtype)

        self.gmean = lib.zeros((self.K, 1), dtype=self.dtype)
        self.gvar  = lib.zeros((self.K, 1), dtype=self.dtype)

        self.grad_beta  = lib.zeros((self.K, 1), dtype=self.dtype)
        self.grad_gamma = lib.zeros((self.K, 1), dtype=self.dtype)

    def fprop(self, fprop_in, scale_weights=0):
        """
        Batch normalization forward pass. Uses a compound kernel call.
        """
        if type(fprop_in) is tuple:
            fprop_in, bsum = fprop_in
        else:
            bsum = None

        if self.fprop_in is None:
            self.fprop_in = fprop_in.reshape(self.dim2)

        if bsum is None:
            self.xsum[:] = self.lib.sum(self.fprop_in, axis=1)
        else:
            self.xsum = bsum

        self.lib.compound_fprop_bn(
            self.fprop_in, self.xsum, self.xvar, self.gmean, self.gvar,
            self.gamma, self.beta, self.fprop_out, self.eps, self.rho, self.relu)

        return self.fprop_out

    def bprop(self, bprop_in, beta=0):

        if self.bprop_in is None:
            self.bprop_in = bprop_in.reshape(self.dim2)

        if self.relu:
            self.bprop_relu(self.bprop_in)

        self.lib.compound_bprop_bn(
            self.bprop_in, self.grad_gamma, self.grad_beta, self.fprop_in,
            self.xsum, self.xvar, self.gamma, self.eps)

        return self.bprop_in

    def fprop_stats(self):
        pass

    def bprop_stats(self):
        pass



# Magic numbers and shift amounts for integer division
# Suitable for when nmax*magic fits in 32 bits
# Shamelessly pulled directly from:
# http://www.hackersdelight.org/hdcodetxt/magicgu.py.txt
def _magic32(nmax, d):
    nc = ((nmax + 1) // d) * d - 1
    nbits = len(bin(nmax)) - 2
    for p in range(0, 2 * nbits + 1):
        if 2 ** p > nc * (d - 1 - (2 ** p - 1) % d):
            m = (2 ** p + d - 1 - (2 ** p - 1) % d) // d
            return (m, p)
    raise ValueError("Can't find magic number for division")

# Magic numbers and shift amounts for integer division
# Suitable for when nmax*magic fits in 64 bits and the shift
# lops off the lower 32 bits
def _magic64(d):
    # 3 is a special case that only ends up in the high bits
    # if the nmax is 0xffffffff
    # we can't use 0xffffffff for all cases as some return a 33 bit
    # magic number
    nmax = 0xffffffff if d == 3 else 0x7fffffff
    magic, shift = _magic32(nmax, d)
    if magic != 1:
        shift -= 32
    return (magic, shift)

# flatten a nested list of lists or values
def _flatten(lst):
    return sum(([x] if not isinstance(x, (list, tuple))
                else _flatten(x) for x in lst), [])

def _ceil_div(x, y):
    return -(-x // y)

@context_dependent_memoize
def _get_sm_count():
    attributes = drv.Context.get_device().get_attributes()
    return attributes[drv.device_attribute.MULTIPROCESSOR_COUNT]



# ----------------------------------------------------------------------------
# Copyright 2015-2016 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
from __future__ import division
from builtins import str

import h5py
import numpy as np


def create_minibatch_x(minibatches, minibatch_markers, epoch_axis):
    """
    Helper function to build x axis for data captured per minibatch.

    Arguments:
        minibatches (int): how many total minibatches
        minibatch_markers (int array): cumulative number of minibatches complete at a given epoch
        epoch_axis (bool): whether to render epoch or minibatch as the integer step in the x axis
    """
    if epoch_axis:
        x = np.zeros((minibatches,))
        last_e = 0
        for e_idx, e in enumerate(minibatch_markers):
            e_minibatches = e - last_e
            x[last_e:e] = e_idx + (np.arange(float(e_minibatches)) / e_minibatches)
            last_e = e
    else:
        x = np.arange(minibatches)

    return x


def create_epoch_x(points, epoch_freq, minibatch_markers, epoch_axis):
    """
    Helper function to build x axis for points captured per epoch.

    Arguments:
        points (int): how many data points need a corresponding x axis points
        epoch_freq (int): are points once an epoch or once every n epochs?
        minibatch_markers (int array): cumulative number of minibatches complete at a given epoch
        epoch_axis (bool): whether to render epoch or minibatch as the integer step in the x axis
    """

    if epoch_axis:
        x = np.zeros((points,))
        last_e = 0
        for e_idx, e in enumerate(minibatch_markers):
            e_minibatches = e - last_e
            if (e_idx + 1) % epoch_freq == 0:
                x[e_idx // epoch_freq] = e_idx + ((e_minibatches - 1) // e_minibatches)
            last_e = e
    else:
        x = minibatch_markers[(epoch_freq - 1)::epoch_freq] - 1

    return x


def h5_cost_data(filename, epoch_axis=True):
    """
    Read cost data from hdf5 file. Generate x axis data for each cost line.

    Arguments:
        filename (str): Filename with hdf5 cost data
        epoch_axis (bool): whether to render epoch or minibatch as the integer step in the x axis

    Returns:
        list of tuples of (name, x data, y data)
    """
    ret = list()
    with h5py.File(filename, "r") as f:

        config, cost, time_markers = [f[x] for x in ['config', 'cost', 'time_markers']]
        total_epochs = config.attrs['total_epochs']
        total_minibatches = config.attrs['total_minibatches']
        minibatch_markers = time_markers['minibatch']

        for name, ydata in cost.items():
            y = ydata[...]
            if ydata.attrs['time_markers'] == 'epoch_freq':
                y_epoch_freq = ydata.attrs['epoch_freq']
                assert len(y) == total_epochs // y_epoch_freq
                x = create_epoch_x(len(y), y_epoch_freq, minibatch_markers, epoch_axis)

            elif ydata.attrs['time_markers'] == 'minibatch':
                assert len(y) == total_minibatches
                x = create_minibatch_x(total_minibatches, minibatch_markers, epoch_axis)

            else:
                raise TypeError('Unsupported data format for h5_cost_data')

            ret.append((name, x, y))

    return ret


def h5_hist_data(filename, epoch_axis=True):
    """
    Read histogram data from hdf5 file. Generate x axis data for each hist line.

    Arguments:
        filename (str): Filename with hdf5 cost data
        epoch_axis (bool): whether to render epoch or minibatch as the integer step in the x axis

    Returns:
        list of tuples of (name, data, dh, dw, bins, offset)
    """
    ret = list()
    with h5py.File(filename, "r") as f:
        if 'hist' in f:
            hists, config = [f[x] for x in ['hist', 'config']]
            bins, offset, time_markers = [hists.attrs[x]
                                          for x in ['bins', 'offset', 'time_markers']]
            total_epochs = config.attrs['total_epochs']
            total_minibatches = config.attrs['total_minibatches']

            for hname, hdata in hists.items():
                dw = total_epochs if (time_markers == 'epoch_freq') else total_minibatches
                dh = bins
                ret.append((hname, hdata[...], dh, dw, bins, offset))

    return ret


def convert_rgb_to_bokehrgba(img_data, downsample=1):
    """
    Convert RGB image to two-dimensional array of RGBA values (encoded as 32-bit integers)
    (required by Bokeh). The functionality is currently not available in Bokeh.
    An issue was raised here: https://github.com/bokeh/bokeh/issues/1699 and this function is a
    modified version of the suggested solution.

    Arguments:
        img_data: img (ndarray, shape: [N, M, 3], dtype: uint8): image data
        dh: height of image
        dw: width of image

    Returns:
        img (ndarray): 2D image array of RGBA values
    """
    if img_data.dtype != np.uint8:
        raise NotImplementedError

    if img_data.ndim != 3:
        raise NotImplementedError

    # downsample for render performance, v-flip since plot origin is bottom left
    # img_data = np.transpose(img_data, (1,2,0))
    img_data = img_data[::-downsample, ::downsample, :]
    img_h, img_w, C = img_data.shape

    # add an alpha channel to the image and recast from pixels of u8u8u8u8 to u32
    bokeh_img = np.dstack([img_data, 255 * np.ones((img_h, img_w), np.uint8)])
    final_image = bokeh_img.reshape(img_h, img_w * (C + 1)).view(np.uint32)

    return final_image


def h5_deconv_data(filename):
    """
    Read deconv visualization data from hdf5 file.

    Arguments:
        filename (str): Filename with hdf5 deconv data

    Returns:
        list of lists. Each inner list represents one layer, and consists of
        tuples (fm, deconv_data)
    """
    ret = list()
    with h5py.File(filename, "r") as f:
        if 'deconv' not in list(f.keys()):
            return None
        act_data = f['deconv/max_act']
        img_data = f['deconv/img']

        for layer in list(act_data.keys()):
            layer_data = list()
            for fm in range(act_data[layer]['vis'].shape[0]):

                # to avoid storing entire dataset, imgs are cached as needed, have to look up
                batch_ind, img_ind = act_data[layer]['batch_img'][fm]
                img_store = img_data['batch_{}'.format(batch_ind)]
                img_cache_ofs = img_store.attrs[str(img_ind)]

                # have to convert from rgb to rgba and cast as uint32 dtype for bokeh
                plot_img = convert_rgb_to_bokehrgba(img_store['HWC_uint8'][:, :, :, img_cache_ofs])
                plot_deconv = convert_rgb_to_bokehrgba(act_data[layer]['vis'][fm])

                layer_data.append((fm, plot_deconv, plot_img))

            ret.append((layer, layer_data))
    return ret



# ----------------------------------------------------------------------------
# Copyright 2014-2016 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""
Defines text datatset preprocessing routines
"""
from future import standard_library
standard_library.install_aliases()  # triggers E402, hence noqa below
from builtins import map  # noqa
import numpy as np  # noqa
import re  # noqa
from neon.util.compat import pickle  # noqa


def clean_string(base):
    """
    Tokenization/string cleaning.
    Original from https://github.com/yoonkim/CNN_sentence/blob/master/process_data.py
    """
    base = re.sub(r"[^A-Za-z0-9(),!?\'\`]", " ", base)
    base = re.sub(r"\'re", " \'re", base)
    base = re.sub(r"\'d", " \'d", base)
    base = re.sub(r"\'ll", " \'ll", base)
    base = re.sub(r"\'s", " \'s", base)
    base = re.sub(r"\'ve", " \'ve", base)
    base = re.sub(r"n\'t", " n\'t", base)
    base = re.sub(r"!", " ! ", base)
    base = re.sub(r",", " , ", base)
    base = re.sub(r"\)", " \) ", base)
    base = re.sub(r"\(", " \( ", base)
    base = re.sub(r"\?", " \? ", base)
    base = re.sub(r"\s{2,}", " ", base)
    return base.strip().lower()


def pad_sentences(sentences, sentence_length=None, dtype=np.int32, pad_val=0.):
    lengths = [len(sent) for sent in sentences]

    nsamples = len(sentences)
    if sentence_length is None:
        sentence_length = np.max(lengths)

    X = (np.ones((nsamples, sentence_length)) * pad_val).astype(dtype=np.int32)
    for i, sent in enumerate(sentences):
        trunc = sent[-sentence_length:]
        X[i, -len(trunc):] = trunc
    return X


def pad_data(path, vocab_size=20000, sentence_length=100, oov=2,
             start=1, index_from=3, seed=113, test_split=0.2):
    f = open(path, 'rb')
    X, y = pickle.load(f)
    f.close()

    np.random.seed(seed)
    np.random.shuffle(X)
    np.random.seed(seed)
    np.random.shuffle(y)

    if start is not None:
        X = [[start] + [w + index_from for w in x] for x in X]
    else:
        X = [[w + index_from for w in x] for x in X]

    if not vocab_size:
        vocab_size = max([max(x) for x in X])

    # by convention, use 2 as OOV word
    # reserve 'index_from' (=3 by default) characters: 0 (padding), 1
    # (start), 2 (OOV)
    if oov is not None:
        X = [[oov if w >= vocab_size else w for w in x] for x in X]

    X_train = X[:int(len(X) * (1 - test_split))]
    y_train = y[:int(len(X) * (1 - test_split))]

    X_test = X[int(len(X) * (1 - test_split)):]
    y_test = y[int(len(X) * (1 - test_split)):]

    X_train = pad_sentences(X_train, sentence_length=sentence_length)
    y_train = np.array(y_train).reshape((len(y_train), 1))

    X_test = pad_sentences(X_test, sentence_length=sentence_length)
    y_test = np.array(y_test).reshape((len(y_test), 1))

    nclass = 1 + max(np.max(y_train), np.max(y_test))

    return (X_train, y_train), (X_test, y_test), nclass


def get_paddedXY(X, y, vocab_size=20000, sentence_length=100, oov=2,
                 start=1, index_from=3, seed=113, shuffle=True):

    if shuffle:
        np.random.seed(seed)
        np.random.shuffle(X)
        np.random.seed(seed)
        np.random.shuffle(y)

    if start is not None:
        X = [[start] + [w + index_from for w in x] for x in X]
    else:
        X = [[w + index_from for w in x] for x in X]

    if not vocab_size:
        vocab_size = max([max(x) for x in X])

    # word ids - pad (0), start (1), oov (2)
    if oov is not None:
        X = [[oov if w >= vocab_size else w for w in x] for x in X]
    else:
        X = [[w for w in x if w < vocab_size] for x in X]

    X = pad_sentences(X, sentence_length=sentence_length)
    y = np.array(y, dtype=np.int32).reshape((len(y), 1))

    return X, y


def get_google_word2vec_W(fname, vocab, vocab_size=1000000, index_from=3):
    """
    Extract the embedding matrix from the given word2vec binary file and use this
    to initalize a new embedding matrix for words found in vocab.

    Conventions are to save indices for pad, oov, etc.:
    index 0: pad
    index 1: oov (or <unk>)
    index 2: <eos>. But often cases, the <eos> has already been in the
    preprocessed data, so no need to save an index for <eos>
    """
    f = open(fname, 'rb')
    header = f.readline()
    vocab1_size, embedding_dim = list(map(int, header.split()))
    binary_len = np.dtype('float32').itemsize * embedding_dim
    vocab_size = min(len(vocab) + index_from, vocab_size)
    W = np.zeros((vocab_size, embedding_dim))

    found_words = {}
    for i, line in enumerate(range(vocab1_size)):
        word = []
        while True:
            ch = f.read(1)
            if ch == b' ':
                word = b''.join(word)
                break
            if ch != '\n':
                word.append(ch)
        if word in vocab:
            wrd_id = vocab[word] + index_from
            if wrd_id < vocab_size:
                W[wrd_id] = np.fromstring(
                    f.read(binary_len), dtype='float32')
                found_words[wrd_id] = 1
        else:
            f.read(binary_len)

    cnt = 0
    for wrd_id in range(vocab_size):
        if wrd_id not in found_words:
            W[wrd_id] = np.random.uniform(-0.25, 0.25, embedding_dim)
            cnt += 1
    assert cnt + len(found_words) == vocab_size

    f.close()

    return W, embedding_dim, vocab_size



# ----------------------------------------------------------------------------
# Copyright 2015-2016 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
from __future__ import division
from neon import NervanaObject
from neon.util.persist import load_class
import logging
import numpy as np

logger = logging.getLogger(__name__)


def get_param_list(layer_list):
    '''
    Returns a flattened list of parameters. Each element in the list
    is a tuple ``((W, dW), states)`` for the parameters ``W``, parameter updates ``dW``,
    and the current set of ``states``.

    Args:
        layer_list (list): List of layers

    Returns:
        param_list (list): List of parameters.
    '''
    plist = []
    for l in layer_list:
        ptuple = l.get_params()
        plist.extend(ptuple) if isinstance(ptuple, list) else plist.append(ptuple)
    return plist


class Optimizer(NervanaObject):

    '''
    The optimizer class handles the gradient update stage of training a neural network.
    Given the current parameters :math:`w`, update parameters
    :math:`\Delta w`, and current state :math:`s`, the optimizer specifies an
    algorithm for performing the update.

    This base class contains to helper functions for scaling the gradients.
    specifices the abstract method optimize, which subclasses should implement. The optimize
    method is called at every minibatch to update the layer parameters.
    '''
    def __init__(self, name=None):
        """
        Class constructor.
        """
        super(Optimizer, self).__init__(name=name)

    def optimize(self, layer_list, epoch):
        """
        Update the parameters for a provided list of layers.

        Args:
            layer_list (list): List of layers to optimize
            epoch (integer): Epoch count of training
        """
        raise NotImplementedError()

    def clip_gradient_norm(self, param_list, clip_norm):
        """
        Returns a scaling factor to apply to the gradients.

        The scaling factor is computed such that the root mean squared
        average of the scaled gradients across all layers will be less than
        or equal to the provided clip_norm value. This factor is always <1, so
        never scales up the gradients.

        Arguments:
            param_list (list): List of layer parameters
            clip_norm (float, optional): Target norm for the gradients. If not provided
                                         the returned scale_factor will equal 1.

        Returns:
            scale_factor (float): Computed scale factor.
        """
        scale_factor = 1
        if clip_norm:
            grad_list = [grad for (param, grad), states in param_list]
            grad_square_sums = sum(self.be.sum(self.be.square(grad)) for grad in grad_list)
            grad_norm = self.be.zeros((1, 1))
            grad_norm[:] = self.be.sqrt(grad_square_sums) / self.be.bsz
            scale_factor = clip_norm / max(float(grad_norm.get()), float(clip_norm))
        return scale_factor

    def clip_value(self, v, abs_bound=None):
        """
        Element-wise clip a gradient or parameter tensor to between
        ``-abs_bound`` and ``+abs_bound``.

        Arguments:
            v (tensor): Tensor of gradients or parameters for a single layer
            abs_bound (float, optional): Value to element-wise clip gradients
                                         or parameters. Defaults to None.

        Returns:
            v (tensor): Tensor of clipped gradients or parameters.

        """
        if abs_bound:
            return self.be.clip(v, -abs(abs_bound), abs(abs_bound))
        else:
            return v


class Schedule(NervanaObject):
    """
    Learning rate schedule.

    By default implements a constant learning rate:

    .. code-block:: python

        # Constant learning rate of 0.01 across training epochs
        optimizer = GradientDescentMomentum(0.01, 0.9, schedule = Schedule())

    Otherwise, the schedule multiplies the learning rate by change at every element in
    ``step_config``.
    For example,

    .. code-block:: python

        schedule = Schedule(step_config=[2, 6], change=0.5)
        optimizer = GradientDescentMomentum(1.0, 0.9, schedule = Schedule())

    will yield a learning rate schedule of:

    .. csv-table::
        :header: "Epoch", "LR"
        :widths: 20, 10

        0, 1.0
        1, 1.0
        2, 0.5
        3, 0.5
        4, 0.5
        5, 0.5
        6, 0.25
        7, 0.25
        8, 0.25
        9, 0.25
    """

    def __init__(self, step_config=None, change=1.):
        """
        Class constructor.

        Arguments:
            step_config (list, optional): Configure the step times (list of epoch indices).
                                          Defaults to None (constant).
            change (int, optional): The learning rate is
                                    multiplied by ``change ** steps``, where ``steps`` is the
                                    number of steps in the step schedule that have passed.
        """

        if isinstance(step_config, list) and isinstance(change, list):
            assert len(step_config) == len(change), "change and step_config must have the same" \
                "length after step_config is deduplicated to do epoch-level LR assignment."

            logger.warn("This functionality will be removed from Schedule in the future. "
                        "Please use the StepSchedule class instead.")

        if isinstance(step_config, int):
            logger.warn("This functionality will be removed from Schedule in the future. "
                        "Please use the PowerSchedule class instead.")

        self.step_config = step_config
        self.change = change
        self.steps = 0

    def get_learning_rate(self, learning_rate, epoch):
        """
        Returns the current learning rate given the epoch and initial learning rate.

        Arguments:
            learning_rate (float): Initial learning rate
            epoch (int): Current epoch, used to calculate the adjusted learning rate

        Returns:
            (float): The adjusted learning rate
        """

        # will be moved to StepSchedule in the future
        if isinstance(self.step_config, list) and isinstance(self.change, list):
            if epoch in self.step_config:
                # steps will store the current lr
                self.steps = self.change[self.step_config.index(epoch)]
            if self.steps == 0:
                return learning_rate
            else:
                return self.steps

        # will be moved to PowerSchedule in the future
        elif isinstance(self.step_config, int):
            self.steps = np.floor(epoch / self.step_config)

        elif isinstance(self.step_config, list):
            self.steps = np.sum(epoch >= np.array(self.step_config))

        return float(learning_rate * self.change ** self.steps)


class StepSchedule(Schedule):
    """
    Steps the learning rate over training time.

    To set a step schedule, pass as arguments ``step_config`` and ``change``. The schedule
    will set the learning rate at ``step[i]`` to ``change[i]``. For example, the call:

    .. code-block:: python

        schedule = Schedule(step_config=[2, 6], change=[0.6, 0.4])

    will set the learning rate to 0.6 at step 2, and to 0.4 at step 6.
    """
    def __init__(self, step_config, change):
        """
        Class constructor.

        Arguments:
            step_config (list): Configure the step times (list of epoch indices)
            change (list): List of learning rates. Must be same length as step_config
        """

        assert isinstance(step_config, list) and isinstance(change, list), \
            "The arguments change and step_config must be lists."

        assert len(step_config) == len(change), \
            "The arguments change and step_config must have the same length."

        self.step_config = step_config
        self.change = change
        self.steps = 0

    def get_learning_rate(self, learning_rate, epoch):
        """
        Returns the current learning rate given the epoch and initial learning rate.

        Arguments:
            learning_rate (float): Initial learning rate
            epoch (int): Current epoch, used to calculate the adjusted learning rate

        Returns:
            (float): The adjusted learning rate
        """
        if epoch in self.step_config:
            # steps will store the current lr
            self.steps = self.change[self.step_config.index(epoch)]
        if self.steps == 0:
            return learning_rate
        else:
            return self.steps


class PowerSchedule(Schedule):
    """
    Multiplies the learning rate by a factor at regular epoch intervals.

    This schedule will multiply the learning rate by
    the factor ``change`` every ``step_config`` epochs. For example,

    .. code-block:: python

        schedule = Schedule(step_config=2, change=0.5)
        optimizer = GradientDescentMomentum(0.1, 0.9, schedule=schedule)

    will yield a learning rate schedule of:

    .. csv-table::
        :header: "Epoch", "LR"
        :widths: 20, 10

        0, 0.1
        1, 0.1
        2, 0.05
        3, 0.05
        4, 0.025
        5, 0.025
        6, 0.0125
        7, 0.0125
    """
    def __init__(self, step_config, change):
        """
        Class constructor.

        Arguments:
            step_config (int): Learning rate update interval (in epochs)
            change (int): Update factor
        """
        assert isinstance(step_config, int), \
            "The argument step_config must be an integer."

        assert not isinstance(change, list), \
            "The argument change must be a float or integer."

        self.step_config = step_config
        self.change = change
        self.steps = 0

    def get_learning_rate(self, learning_rate, epoch):
        """
        Returns the current learning rate given the epoch and initial learning rate.

        Arguments:
            learning_rate (float): Initial learning rate
            epoch (int): Current epoch, used to calculate the adjusted learning rate.

        Returns:
            (float): The adjusted learning rate.
        """
        self.steps = np.floor(epoch / self.step_config)

        return float(learning_rate * self.change ** self.steps)


class ExpSchedule(Schedule):
    """
    Exponential learning rate schedule. This schedule implements

    .. math::
        \\alpha(t) = \\frac{\\alpha_\\circ}{1 + \\beta t}

    where :math:`\\beta` is the decay rate, and :math:`\\alpha_\\circ` is the
    initial learning rate.
    """
    def __init__(self, decay):
        """
        Class constructor.

        Arguments:
            decay (float): Decay rate.
        """
        self.decay = decay

    def get_learning_rate(self, learning_rate, epoch):
        """
        Returns the current learning rate given the epoch and initial learning rate.

        Arguments:
            learning_rate (float): Initial learning rate
            epoch (int): Current epoch, used to calculate the adjusted learning rate.

        Returns:
            (float): The adjusted learning rate.
        """
        return float(learning_rate / (1. + self.decay * epoch))


class PolySchedule(Schedule):
    """
    Polynomial learning rate schedule.

    This schedule takes as input the total number of epochs :math:`T` and a power :math:`\\beta`,
    and produces the learning schedule:

    .. math::
        \\alpha(t) = \\alpha_\\circ \\times\\left(1-\\frac{t}{T}\\right)^\\beta

    where :math:`\\alpha_\\circ` is the initial learning rate.
    """

    def __init__(self, total_epochs, power):
        """
        Class constructor.

        Arguments:
            total_epochs (int): Total number of epochs over which to calculate interpolated decay
            power (float): Total decay parameter
        """
        self.total_epochs = np.float32(total_epochs)
        self.power = power

    def get_learning_rate(self, learning_rate, epoch):
        """
        Returns the current learning rate given the epoch and initial learning rate.

        Arguments:
            learning_rate (float): Initial learning rate
            epoch (int): Current epoch, used to calculate the adjusted learning rate.

        Returns:
            (float): The adjusted learning rate.
        """
        return float(learning_rate * (1. - epoch // self.total_epochs) ** self.power)


class ShiftSchedule(Schedule):
    """
    Binary shift learning rate schedule.

    Arguments:
        interval (int): interval in epochs the learning rate is shifted
        shift_size (int): amount to shift
    """
    def __init__(self, interval, shift_size=1):
        self.interval = interval
        self.shift_size = shift_size

    def get_learning_rate(self, learning_rate, epoch):
        total_shift = -1 * self.shift_size * int(epoch/self.interval)
        return float(self.be.shift(learning_rate, total_shift, value=False).get())


class GradientDescentMomentum(Optimizer):

    """
    Stochastic gradient descent with momentum.

    Given the parameters :math:`\\theta`, the learning rate :math:`\\alpha`,
    and the gradients :math:`\\nabla J(\\theta; x)`
    computed on the minibatch data :math:`x`, SGD updates the parameters via

    .. math::
        \\theta' = \\theta - \\alpha\\nabla J(\\theta; x)

    Here we implement SGD with momentum. Momentum tracks the history of
    gradient updates to help the system move faster through saddle points.
    Given the additional parameters: momentum :math:`\gamma`, weight decay :math:`\lambda`,
    and current velocity :math:`v`, we use the following update equations

    .. math::
        v' = \\gamma v - \\alpha(\\nabla J(\\theta; x) + \\lambda\\theta)
        theta' = \\theta + v'

    The optional `nesterov` parameter implements Nesterov Accelerated Gradient.
    If this is set, we use the following update equations instead
    .. math::
        v' = \\gamma^2 v + \\alpha (\\gamma + 1) (\\nabla J(\\theta; x) + \\lambda\\theta)
        theta' = \\theta + v'

    Example usage:

    .. code-block:: python

        from neon.optimizers import GradientDescentMomentum

        # use SGD with learning rate 0.01 and momentum 0.9, while
        # clipping the gradient magnitude to between -5 and 5.
        opt = GradientDescentMomentum(0.01, 0.9, gradient_clip_value = 5)
    """

    def __init__(self, learning_rate, momentum_coef, stochastic_round=False,
                 wdecay=0.0, gradient_clip_norm=None, gradient_clip_value=None,
                 param_clip_value=None, name=None, schedule=Schedule(),
                 nesterov=False):
        """
        Class constructor.

        Arguments:
            learning_rate (float): Multiplicative coefficient of updates
            momentum_coef (float): Coefficient of momentum
            stochastic_round (bool, optional): Set this to True for stochastic
                                               rounding.  If False (default)
                                               rounding will be to nearest.  If
                                               True use default width
                                               stochastic rounding.  Note that
                                               this only affects the GPU
                                               backend.
            wdecay (float, optional): Amount of weight decay.  Defaults to 0
            gradient_clip_norm (float, optional): Target gradient norm.
                                                  Defaults to None.
            gradient_clip_value (float, optional): Value to element-wise clip
                                                   gradients.
                                                   Defaults to None.
            param_clip_value (float, optional): Value to element-wise clip
                                                parameters.
                                                Defaults to None.
            name (str, optional): the optimizer's layer's pretty-print name.
                                  Defaults to "gdm".
            schedule (neon.optimizers.optimizer.Schedule, optional): Learning
                        rate schedule.  Defaults to a constant learning rate.
            nesterov (bool, optional): Use nesterov accelerated gradient.
                                       Defaults to False.
        """
        super(GradientDescentMomentum, self).__init__(name=name)
        self.learning_rate, self.momentum_coef = (learning_rate, momentum_coef)
        self.gradient_clip_norm = gradient_clip_norm
        self.gradient_clip_value = gradient_clip_value
        self.param_clip_value = param_clip_value
        self.wdecay = wdecay
        self.schedule = schedule
        self.stochastic_round = stochastic_round
        self.nesterov = nesterov
        if self.momentum_coef == 0 and self.nesterov:
            raise ValueError("nesterov requires non-zero momentum")

    def optimize(self, layer_list, epoch):
        """
        Apply the learning rule to all the layers and update the states.

        Arguments:
            layer_list (list): a list of Layer objects to optimize.
            epoch (int): the current epoch, needed for the Schedule object.
        """
        lrate = self.schedule.get_learning_rate(self.learning_rate, epoch)
        param_list = get_param_list(layer_list)

        scale_factor = self.clip_gradient_norm(param_list, self.gradient_clip_norm)

        for (param, grad), states in param_list:
            param.rounding = self.stochastic_round
            if len(states) == 0 and self.momentum_coef != 0:
                states.append(self.be.zeros_like(grad))

            grad = grad / self.be.bsz
            grad = self.clip_value(grad, self.gradient_clip_value)

            if self.momentum_coef == 0:
                param[:] = (- lrate * scale_factor) * grad +\
                           (1 - lrate * self.wdecay) * param
                param = self.clip_value(param, self.param_clip_value)
            else:
                grad = scale_factor * grad + self.wdecay * param
                velocity = states[0]
                velocity[:] = self.momentum_coef * velocity - lrate * grad

                # Nesterov accelerated gradient (NAG) is implemented the same
                # as in torch's "sgd.lua". It's a reformulation of Sutskever's
                # NAG equation found in "On the importance of initialization
                # and momentum in deep learning".
                if self.nesterov:
                    param[:] = self.clip_value(
                               param + self.momentum_coef * velocity
                               - lrate * grad, self.param_clip_value)
                else:
                    param[:] = self.clip_value(
                                param + velocity, self.param_clip_value)


class RMSProp(Optimizer):

    """
    Root Mean Square propagation.

    Root Mean Square (RMS) propagation protects against vanishing and
    exploding gradients. In RMSprop, the gradient is divided by a running
    average of recent gradients. Given the parameters :math:`\\theta`, gradient :math:`\\nabla J`,
    we keep a running average :math:`\\mu` of the last :math:`1/\\lambda` gradients squared.
    The update equations are then given by

    .. math::

        \\mu' &= \\lambda\\mu + (1-\\lambda)(\\nabla J)^2

    .. math::

        \\theta' &= \\theta - \\frac{\\alpha}{\\sqrt{\\mu + \\epsilon} + \\epsilon}\\nabla J

    where we use :math:`\\epsilon` as a (small) smoothing factor to prevent from dividing by zero.
    """

    def __init__(self, stochastic_round=False, decay_rate=0.95, learning_rate=2e-3, epsilon=1e-6,
                 gradient_clip_norm=None, gradient_clip_value=None, param_clip_value=None,
                 name=None, schedule=Schedule()):
        """
        Class constructor.

        Arguments:
            stochastic_round (bool): Set this to True for stochastic rounding.
                                     If False rounding will be to nearest.
                                     If True will perform stochastic rounding using default width.
                                     Only affects the gpu backend.
            decay_rate (float): decay rate of states
            learning_rate (float): the multiplication coefficent of updates
            epsilon (float): smoothing epsilon to avoid divide by zeros
            gradient_clip_norm (float, optional): Target gradient norm.
                                                  Defaults to None.
            gradient_clip_value (float, optional): Value to element-wise clip
                                                   gradients.
                                                   Defaults to None.
            param_clip_value (float, optional): Value to element-wise clip
                                                parameters.
                                                Defaults to None.
            schedule (neon.optimizers.optimizer.Schedule, optional): Learning rate schedule.
                                                                     Defaults to a constant.
        Notes:
            Only constant learning rate is supported currently.
        """
        super(RMSProp, self).__init__(name=name)
        self.state_list = None

        self.epsilon = epsilon
        self.decay_rate = decay_rate
        self.learning_rate = learning_rate
        self.schedule = schedule
        self.gradient_clip_norm = gradient_clip_norm
        self.gradient_clip_value = gradient_clip_value
        self.param_clip_value = param_clip_value
        self.stochastic_round = stochastic_round

    def optimize(self, layer_list, epoch):
        """
        Apply the learning rule to all the layers and update the states.

        Arguments:
            layer_list (list): a list of Layer objects to optimize.
            epoch (int): the current epoch, needed for the Schedule object.
        """
        lrate = self.schedule.get_learning_rate(self.learning_rate, epoch)
        epsilon, decay = (self.epsilon, self.decay_rate)
        param_list = get_param_list(layer_list)

        scale_factor = self.clip_gradient_norm(param_list, self.gradient_clip_norm)

        for (param, grad), states in param_list:

            param.rounding = self.stochastic_round
            if len(states) == 0:
                states.append(self.be.zeros_like(grad))

            grad = grad / self.be.bsz
            grad = self.clip_value(grad, self.gradient_clip_value)

            # update state
            state = states[0]
            state[:] = decay * state + self.be.square(grad) * (1.0 - decay)

            param[:] = self.clip_value(
                        param - (scale_factor * grad * lrate)
                        / (self.be.sqrt(state + epsilon) + epsilon),
                        self.param_clip_value)


class Adagrad(Optimizer):

    """
    Adagrad optimization algorithm.

    Adagrad is an algorithm that adapts the learning rate individually for each parameter
    by dividing by the :math:`L_2`-norm of all previous gradients. Given the parameters
    :math:`\\theta`, gradient :math:`\\nabla J`, accumulating norm :math:`G`, and smoothing
    factor :math:`\\epsilon`, we use the update equations:

    .. math::

        G' = G + (\\nabla J)^2

    .. math::

        \\theta' = \\theta - \\frac{\\alpha}{\sqrt{G' + \\epsilon}} \\nabla J

    where the smoothing factor :math:`\\epsilon` prevents from dividing by zero.
    By adjusting the learning rate individually for each parameter, Adagrad adapts
    to the geometry of the error surface. Differently scaled weights have appropriately scaled
    update steps.

    Example usage:

    .. code-block:: python

        from neon.optimizers import Adagrad

        # use Adagrad with a learning rate of 0.01
        optimizer = Adagrad(learning_rate=0.01, epsilon=1e-6)

    """

    def __init__(self, stochastic_round=False, learning_rate=0.01, epsilon=1e-6,
                 gradient_clip_norm=None, gradient_clip_value=None,
                 param_clip_value=None, name=None):
        """
        Class constructor.

        Arguments:
            stochastic_round (bool): Set this to True for stochastic rounding.
                                     If False rounding will be to nearest.
                                     If True will perform stochastic rounding using default width.
                                     Only affects the gpu backend.
            learning_rate (float): the multiplication coefficent of updates
            epsilon (float): smoothing epsilon to avoid divide by zeros
            gradient_clip_norm (float, optional): Target gradient norm.
                                                  Defaults to None.
            gradient_clip_value (float, optional): Value to element-wise clip
                                                   gradients.
                                                   Defaults to None.
            param_clip_value (float, optional): Value to element-wise clip
                                                parameters.
                                                Defaults to None.
        Notes:
            Only constant learning rate is supported currently.
        """
        super(Adagrad, self).__init__(name=name)
        self.state_list = None
        self.epsilon = epsilon
        self.learning_rate = learning_rate
        self.gradient_clip_norm = gradient_clip_norm
        self.gradient_clip_value = gradient_clip_value
        self.param_clip_value = param_clip_value
        self.stochastic_round = stochastic_round

    def optimize(self, layer_list, epoch):
        """
        Apply the learning rule to all the layers and update the states.

        Arguments:
            layer_list (list): a list of Layer objects to optimize.
            epoch (int): the current epoch, needed for the Schedule object.
        """
        lrate, epsilon = (self.learning_rate, self.epsilon)
        param_list = get_param_list(layer_list)

        scale_factor = self.clip_gradient_norm(param_list, self.gradient_clip_norm)

        for (param, grad), states in param_list:

            param.rounding = self.stochastic_round
            if len(states) == 0:
                states.append(self.be.zeros_like(grad))

            grad = grad / self.be.bsz
            grad = self.clip_value(grad, self.gradient_clip_value)

            # update state
            state = states[0]
            state[:] = state + self.be.square(grad)
            param[:] = self.clip_value(
                        param - (scale_factor * grad * lrate)
                        / (self.be.sqrt(state + epsilon)), self.param_clip_value)


class Adadelta(Optimizer):

    """
    Adadelta optimization algorithm.

    Similar to RMSprop, Adadelta tracks the running average of the
    gradients, :math:`\\mu_J`, over a window size :math:`1/\\lambda`, where
    :math:`\\lambda` is the parameter ``decay``. Adadelta also tracks an average of the
    recent update steps, which we denote as :math:`\\mu_\\theta`, and sets the learning rate
    as the ratio of the two averages:

    .. math::

        \\mu_J' &= \\lambda\\mu_J + (1-\\lambda) (\\nabla J)^2

    .. math::

        \\Delta \\theta &= \\sqrt{\\frac{\\mu_\\theta + \\epsilon}{\\mu_J' + \\epsilon}} \\nabla J

    .. math::

        \\mu_\\theta &= \\lambda \\mu_\\theta + (1-\\rho) (\\Delta \\theta)^2

    .. math::

        \\theta &= \\theta - \\Delta \\theta

    Note that the learning rate is a ratio of the average updates from the
    previous step, :math:`\\mu_\\theta`, divided by the average gradients including the current
    step, :math:`\\mu'_J`.

    Example usage:

    .. code-block:: python

        from neon.optimizers import Adadelta
        # use Adagrad with a learning rate of 0.01
        optimizer = Adadelta(decay=0.95, epsilon=1e-6)

    """

    def __init__(self, stochastic_round=False, decay=0.95, epsilon=1e-6,
                 param_clip_value=None, name=None):
        """
        Class constructor.

        Args:
            stochastic_round (bool): Set this to True for stochastic rounding.
                                     If False rounding will be to nearest.
                                     If True will perform stochastic rounding using default width.
                                     Only affects the gpu backend.
            decay: decay parameter in Adadelta
            epsilon: epsilon parameter in Adadelta
            param_clip_value (float, optional): Value to element-wise clip
                                                parameters.
                                                Defaults to None.
        """
        super(Adadelta, self).__init__(name=name)
        self.decay = decay
        self.epsilon = epsilon
        self.stochastic_round = stochastic_round
        self.param_clip_value = param_clip_value

    def optimize(self, layer_list, epoch):
        """
        Apply the learning rule to all the layers and update the states.

        Arguments:
            param_list (list): a list of tuples of the form ((param, grad), state),
                               corresponding to parameters, grads,
                               and states of layers to be updated
            epoch (int): the current epoch, needed for the Schedule object.
        """
        epsilon, decay = (self.epsilon, self.decay)

        param_list = get_param_list(layer_list)

        for (param, grad), states in param_list:
            param.rounding = self.stochastic_round

            if len(states) == 0:
                # E[Grad^2], E[Delt^2], updates
                states.extend([self.be.zeros_like(grad) for i in range(3)])

            grad = grad / self.be.bsz
            states[0][:] = states[0] * decay + (1. - decay) * grad * grad
            states[2][:] = self.be.sqrt((states[1] + epsilon) / (states[0] + epsilon)) * grad
            states[1][:] = states[1] * decay + (1. - decay) * states[2] * states[2]

            param[:] = self.clip_value(param - states[2], self.param_clip_value)


class Adam(Optimizer):

    """
    Adam optimizer.

    The Adam optimizer combines features from RMSprop and Adagrad. We
    accumulate both the first and second moments of the gradient with decay
    rates :math:`\\beta_1` and :math:`\\beta_2` corresponding to window sizes of
    :math:`1/\\beta_1` and :math:`1/\\beta_2`, respectively.

    .. math::
        m' &= \\beta_1 m + (1-\\beta_1) \\nabla J

    .. math::
        v' &= \\beta_2 v + (1-\\beta_2) (\\nabla J)^2

    We update the parameters by the ratio of the two moments:

    .. math::
        \\theta = \\theta - \\alpha \\frac{\\hat{m}'}{\\sqrt{\\hat{v}'}+\\epsilon}

    where we compute the bias-corrected moments :math:`\\hat{m}'` and :math:`\\hat{v}'` via

    .. math::
        \\hat{m}' &= m'/(1-\\beta_1^t)

    .. math::
        \\hat{v}' &= v'/(1-\\beta_1^t)

    Example usage:

    .. code-block:: python

        from neon.optimizers import Adam

        # use Adam
        optimizer = Adam(learning_rate=0.001, beta_1=0.9, beta_2=0.999)

    """

    def __init__(self, stochastic_round=False, learning_rate=0.001, beta_1=0.9, beta_2=0.999,
                 epsilon=1e-8, gradient_clip_norm=None, gradient_clip_value=None,
                 param_clip_value=None, name="adam"):
        """
        Class constructor.

        Args:
            stochastic_round (bool): Set this to True for stochastic rounding.
                                     If False rounding will be to nearest.
                                     If True will perform stochastic rounding using default width.
                                     Only affects the gpu backend.
            learning_rate (float): the multiplicative coefficient of updates
            beta_1 (float): Adam parameter beta1
            beta_2 (float): Adam parameter beta2
            epsilon (float): numerical stability parameter
            gradient_clip_norm (float, optional): Target gradient norm.
                                                  Defaults to None.
            gradient_clip_value (float, optional): Value to element-wise clip gradients.
                                                   Defaults to None.
            param_clip_value (float, optional): Value to element-wise clip parameters.
                                                Defaults to None.
        """
        super(Adam, self).__init__(name=name)
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.epsilon = epsilon
        self.learning_rate = learning_rate
        self.stochastic_round = stochastic_round
        self.gradient_clip_norm = gradient_clip_norm
        self.gradient_clip_value = gradient_clip_value
        self.param_clip_value = param_clip_value
        self.t = 0

    def optimize(self, layer_list, epoch):
        """
        Apply the learning rule to all the layers and update the states.

        Arguments:
            param_list (list): a list of tuples of the form ((param, grad), state),
                corresponding to parameters, grads, and states of layers to be updated
            epoch (int): the current epoch, needed for the Schedule object.
        """
        self.t = self.t + 1
        l = (self.learning_rate * self.be.sqrt(1 - self.beta_2 ** self.t) /
             (1 - self.beta_1 ** self.t))

        param_list = get_param_list(layer_list)

        scale_factor = self.clip_gradient_norm(param_list, self.gradient_clip_norm)

        for (param, grad), states in param_list:
            param.rounding = self.stochastic_round
            if len(states) == 0:
                # running_1st_mom, running_2nd_mom
                states.extend([self.be.zeros_like(grad) for i in range(2)])

            grad = grad / self.be.bsz
            grad = self.clip_value(grad, self.gradient_clip_value)

            m, v = states
            m[:] = m * self.beta_1 + (1. - self.beta_1) * grad
            v[:] = v * self.beta_2 + (1. - self.beta_2) * grad * grad

            param[:] = self.clip_value(
                        param - (scale_factor * l * m)
                        / (self.be.sqrt(v) + self.epsilon), self.param_clip_value)


class ShiftAdaMax(Optimizer):

    """
    Shift based AdaMax. http://arxiv.org/pdf/1602.02830v3.pdf
    """

    def __init__(self, stochastic_round=False, learning_rate=0.002, beta_1=0.9, beta_2=0.999,
                 epsilon=1e-8, schedule=Schedule(), name="ShiftAdaMax"):
        """
        Args:
            stochastic_round (bool): Set this to True for stochastic rounding.
                                     If False rounding will be to nearest.
                                     If True will perform stochastic rounding using default width.
                                     Only affects the gpu backend.
            learning_rate (float): the multiplicative coefficient of updates
            beta_1 (float): Adam parameter beta1
            beta_2 (float): Adam parameter beta2
            epsilon (float): numerical stability parameter
            schedule (neon.optimizers.optimizer.Schedule, optional): Learning rate schedule.
                                                                     Defaults to a constant.
        """
        super(ShiftAdaMax, self).__init__(name=name)
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.epsilon = epsilon
        self.learning_rate = learning_rate
        self.stochastic_round = stochastic_round
        self.schedule = schedule

    def optimize(self, layer_list, epoch):
        """
        Apply the learning rule to all the layers and update the states.

        Arguments:
            param_list (list): a list of tuples of the form ((param, grad), state),
                corresponding to parameters, grads, and states of layers to be updated
            epoch (int): the current epoch, needed for the Schedule object.
        """
        t = epoch + 1
        lrate = self.schedule.get_learning_rate(self.learning_rate, epoch)
        l = lrate / (1 - self.beta_1 ** t)

        param_list = get_param_list(layer_list)

        for (param, grad), states in param_list:
            param.rounding = self.stochastic_round
            if len(states) == 0:
                # running_1st_mom, running_2nd_mom
                states.extend([self.be.zeros_like(grad) for i in range(3)])

            grad = grad / self.be.bsz
            m, v, inv_v = states
            m[:] = m * self.beta_1 + (1. - self.beta_1) * grad
            v[:] = self.be.maximum(v * self.beta_2, self.be.absolute(grad))

            inv_v[:] = 1.0 / (v + self.epsilon)
            param[:] = param - self.be.shift(self.be.shift(m, inv_v), l)
            self.be.clip(param, -1, 1, param)


class MultiOptimizer(Optimizer):

    """
    A wrapper class for using multiple Optimizers within the same model.

    To assign different optimizers to different layers we first define
    the different optimizers:

    .. code-block:: python

        from neon.optimizers import GradientDescentMomentum, RMSprop

        optimizer_A = GradientDescentMomentum(learning_rate=0.01, momentum_coef=0.9)
        optimizer_B = GradientDescentMomentum(learning_rate=0.05, momentum_coef=0.9)
        optimizer_C = RMSprop(learning_rate=2e-3, decay_rate=0.95)

    Then, we instantiate this class and pass a
    dictionary mapping layers to optimizers. The keys can either be:
    ``default``, a layer class name (e.g. ``Bias``), or the Layer's name
    attribute. The latter takes precedence for finer layer-to-layer control.

    For example, if we have the following layers,

    .. code-block:: python

        layers = []
        layers.append(Linear(nout = 100, init=Gaussian(), name="layer_one"))
        layers.append(Linear(nout = 50, init=Gaussian(), name="layer_two"))
        layers.append(Affine(nout = 5, init=Gaussian(), activation=Softmax()))

    we can define multiple optimizers with

    .. code-block:: python

        from neon.optimizers import MultiOptimizer

        # dictionary of mappings
        mapping = {'default': optimizer_A, # default optimizer
                   'Linear': optimizer_B, # all layers from the Linear class
                   'layer_two': optimizer_C} # this overrides the previous entry

        # use multiple optimizers
        opt = MultiOptimizer(mapping)

    After definition, we have the following mapping

    +----------------------+----------------------------+
    | Layer                | Optimizer                  |
    +======================+============================+
    | ``layer_one``        | ``optimizer_B``            |
    +----------------------+----------------------------+
    | ``layer_two``        | ``optimizer_C``            |
    +----------------------+----------------------------+
    | ``Affine.Linear``    | ``optimizer_B``            |
    +----------------------+----------------------------+
    | ``Affine.Bias``      | ``optimizer_A``            |
    +----------------------+----------------------------+
    | ``Affine.Softmax``   | ``None (no parameters)``   |
    +----------------------+----------------------------+
    """

    def __init__(self, optimizer_mapping, name=None):
        """
        Class constructor.

        Args:
            optimizer_mapping (dict): dictionary specifying the mapping of layers to optimizers.
                Key: ``'default'``, layer class name or layer `name` attribute.
                Don't name your layers ``'default'``. Value: the optimizer object to
                use for those layers.
        """
        super(MultiOptimizer, self).__init__(name=name)
        self.optimizer_mapping = optimizer_mapping
        assert 'default' in self.optimizer_mapping, "Must specify a default" \
            "optimizer in layer type to optimizer mapping"

        self.map_list = None
        self.map_list_cache = dict()

    @classmethod
    def gen_class(cls, pdict):
        for key in pdict['optimizer_mapping']:
            # these should be optimizers
            typ = pdict['optimizer_mapping'][key]['type']
            ocls = load_class(typ)
            if 'config' not in pdict['optimizer_mapping'][key]:
                pdict['optimizer_mapping'][key]['config'] = {}
            conf = pdict['optimizer_mapping'][key]['config']
            pdict['optimizer_mapping'][key] = ocls.gen_class(conf)
        return cls(**pdict)

    def get_description(self):
        desc = {'type': self.modulenm}
        desc['config'] = {'optimizer_mapping': {}}
        for key in self.optimizer_mapping:
            opt_desc = self.optimizer_mapping[key].get_description()
            desc['config']['optimizer_mapping'][key] = opt_desc
        return desc

    def _map_optimizers(self, layer_list):
        """
        maps the optimizers to their corresponding layers
        """
        map_list = dict()
        for layer in layer_list:
            classname = layer.__class__.__name__
            name = layer.name
            opt = None
            if name in self.optimizer_mapping:
                opt = self.optimizer_mapping[name]
            elif classname in self.optimizer_mapping:
                opt = self.optimizer_mapping[classname]
            else:
                opt = self.optimizer_mapping['default']

            if opt not in map_list:
                map_list[opt] = [layer]
            else:
                map_list[opt].append(layer)
        return map_list

    def _reset_mapping(self, new_mapping):
        """
        Pass this optimizer a new mapping, and on subsequent optimize call, the
        mapping will be refreshed (since map_list will be recreated)
        """
        self.optimizer_mapping = new_mapping
        self.map_list = None

    def optimize(self, layer_list, epoch):
        """
        Determine which optimizer in the container should go with which layers,
        then apply their optimize functions to those layers.

        Notes:

        We can recalculate ``map_list`` in case ``optimizer_mapping`` changes
        during training.
        """

        if not id(layer_list) in self.map_list_cache:
            self.map_list = self._map_optimizers(layer_list)
            self.map_list_cache[id(layer_list)] = self.map_list
        else:
            self.map_list = self.map_list_cache[id(layer_list)]

        for opt in self.map_list:
            opt.optimize(self.map_list[opt], epoch)



# ----------------------------------------------------------------------------
# Copyright 2015-2016 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
from neon.layers.layer import (Linear, Bias, Affine, Conv, Convolution, GeneralizedCost, Dropout,
                               Pooling, Activation, DataTransform, BatchNorm, BatchNormAutodiff,
                               ShiftBatchNorm, Deconv, Deconvolution, GeneralizedCostMask, LookupTable,
                               BranchNode, SkipNode, LRN, BinaryAffine, BinaryLinear, Reshape,
                               RoiPooling, GeneralizedGANCost)
from neon.layers.recurrent import (Recurrent, LSTM, GRU, RecurrentSum, RecurrentMean, RecurrentLast,
                                   BiRNN, BiBNRNN, BiLSTM, DeepBiRNN, DeepBiLSTM)
from neon.layers.container import (Tree, Sequential, MergeMultistream, MergeBroadcast, Multicost,
                                   MergeSum, SingleOutputTree, Seq2Seq, SkipThought)
